#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""HARPIA Microscopy Module Python library.

This script is a part of the HARPIA Microscopy Kit, which is a set of tools for
the alignment, characterization and troubleshooting of the HARPIA-MM extension
of the HARPIA spectroscopy system.

support@lightcon.com
www.lightcon.com
All rights reserved.
© 2019-2024 Light Conversion
"""
import numpy as np
from lkcom.image import make_gif, load_img, comb_img

def enumerate_overlap_scan_files(path=".", **kwargs):
    """Enumerate pump-probe overlap data files.

    File names are returned sorted by Z positions in um.
    """
    [file_names, zarr] = enum_files(
        path=path, extension="png", preffix="img", **kwargs)
    zarr = zarr*1000
    return [file_names, zarr]

def get_number_of_pixels(sensor):
    """Get the number of camera pixels.

    This helper function is used to make sure the number of camera pixels is
    correct. Physical calibration errors might creep in if the number of pixels
    is hardcoded and especially if the camera sensor is changed.

    The get_camera_pixel_size() function serves a similar purpose.
    """
    if sensor=="ICX445":
        return [1288, 964]

    print("Unknown sensor " + str(sensor))

def get_pixel_size(xarr, yarr):
    """Get mean pixel size of XY coordinate arrays."""
    return np.mean(np.concatenate((np.diff(xarr), np.diff(yarr))))

def gen_combined_focusing_anim(verbose=False):
    """Generate a combined focusing animation showing pump and probe beams.

    TODO: Work in progress.
    """
    path_pu = r"Pump no Cube"
    path_pr = r"Probe no Cube"

    print("Enumerating pump files...")
    [file_names_pu, Z_arr_pu] = enum_files(path=path_pu, extension="dat", preffix="I", files_to_remove=".\\I.dat", prepend_path=True, verbose=verbose)
    print("Enumerating probe files...")
    [file_names_pr, Z_arr_pr] = enum_files(path=path_pr, extension="dat", preffix="I", files_to_remove=".\\I.dat", prepend_path=True, verbose=verbose)

    print("Reading pump images...")
    images_pu = load_img(file_names=file_names_pu)
    print("Reading probe images...")
    images_pr = load_img(file_names=file_names_pr)

    images = comb_img(images_pu, images_pr)

    labels = []
    for Z in Z_arr_pr:
        labels.append("Z={:+d}".format(int(np.round(Z/10)*10)))

    make_gif(images=images, labels=labels, resize=True, verbose=verbose)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Pump-probe XYZ overlap analysis.

Generate a pump-probe XYZ overlap report from data obtained using the
run_cam_focus_can.py script. The overlap data is obtained by moving a camera
through the focus using a sample. The report contains three panels which show
the FWHM beam size and the focus X and Y position as a function of the Z
position.

Please specify the objective type and the device serial number when prompted.

 - A note on pinhole overlap scan -
The HARPIA Microscopy Module was first aligned using a pinhole transmission
scan in which the transmitted power of pump and probe beams through a pinhole
was recorded as the pinhole was scanned around. A faster camera-based method
was developed later and the pinhole method was no longer used. The analysis
script diverged with limited support for pinhole scans in several places by
using the mode='pinhole' argument. Actual pinhole scan data analysis was not
tested in some time and likely got broken somewhere along the way. Furthermore,
a better version of the pinhole scan was developed for spectral PSF spectral
analysis, making the pinhole scan mode for this code deprecated.

Since the pinhole overlap scan method was developed first, it is often referred
to as 'OverlapScan', 'CalibrationScan', or 'XYZScan' with no reference of it
being specific to the pinhole scan only.

Code relating to pinhole scan analysis was removed from this module on
November 2023. It can still be found in the harpiamm code repo if needed.


This script is a part of the HARPIA Microscopy Kit, which is a set of tools for
the alignment, characterization and troubleshooting of the HARPIA-MM extension
of the HARPIA spectroscopy system.

support@lightcon.com
www.lightcon.com
All rights reserved.
© 2019-2024 Light Conversion
"""
import os
import time
from warnings import warn
from pathlib import Path

import numpy as np
import matplotlib.pyplot as plt

from scipy.ndimage.measurements import center_of_mass
from PIL import Image

from lkcom.util import cut_by_x_range, get_common_range, isnone, \
     find_closest, get_color, get_exception_info_str, LogPrinter
from lkcom.standard_func import poly_2
from lkcom.string import join_paths
from lkcom.dataio import enum_files, check_file_exists
from lkcom.image import make_gif
from lkcom.plot import add_x_marker, add_y_marker, export_figure

from lkfit.gaussian_fit import fit_gaussian_2d
from lkfit.poly import fit_poly_2, get_poly_2_max_x

from harpiamm.harpiamm_ui import get_report_metainfo


def get_default_camera_sensor():
    """Get default camera sensor name."""
    return 'ICX445'

def get_camera_pixel_size(sensor):
    """Get the physical size of camera pixels.

    This helper function is used to make sure the camera pixel size is correct.
    Physical calibration errors might creep in if the pixel size is hardcoded
    and especially if the camera sensor is changed.

    The get_number_of_pixels() function serves a similar purpose.
    """
    if sensor=="ICX445":
        return 3.8

    print("Unknown sensor " + sensor)
    return None

def get_default_overlap_fit_result_filename():
    """Get default file name for overlap fit results."""
    return "fit_arr.txt"

def get_xyz_overlap_parameters(path='.', verbose=False, **kwargs):
    """Get XYZ overlap parameters.

    The retrieved parameters for every XY image are: Z position, controid
    locations, beam widths and maximum intensity.
    """
    cam_sensor = kwargs.get('cam_sensor')
    if cam_sensor is None:
        cam_sensor = get_default_camera_sensor()
        print("WARNING: camera sensor not defined, assuming {}".format(
            cam_sensor))

    # Load stored fit parameters, if available
    fit_arr_filename = join_paths(
        path, get_default_overlap_fit_result_filename())

    if check_file_exists(fit_arr_filename):
        fit_arr = np.loadtxt(fit_arr_filename)
        zarr = fit_arr[:, 0]
        ampl = fit_arr[:, 1]
        center = fit_arr[:, 2:4]
        width = fit_arr[:, 4:6]
        print("Loading stored fit parameters")
        return [zarr, ampl, center, width]

    # If no stored parameters found, enumerate, load and fit every image
    [file_names, zarr] = enum_files(
        path=path, extension="png", preffix="img", prepend_path=False,
        verbose=verbose, neg_inds=False)

    # Sort data so that Z position is increasing
    sort_order = np.argsort(zarr)
    zarr = zarr[sort_order]
    file_names = [file_names[i] for i in sort_order]

    pxsz = get_camera_pixel_size(cam_sensor)

    orig_path = os.getcwd()
    try:
        os.chdir(path)

        numf = len(file_names)

        ampl = np.ndarray(numf)
        center = np.ndarray([numf, 2])
        width = np.ndarray([numf, 2])

        fit_arr = np.ndarray([len(file_names), 8])
        fit_arr.fill(np.nan)


        # Pump-probe overlap images from a sample plane camera capture way more
        # area than is needed for overlap characterization. Focused beams are
        # just a few pixels across, so it makes sense to crop the image to
        # a smaller size to increase fitting speed. The cropped image size is
        # determined from the estimated std. dev. of the 2D gaussian fit so
        # there are many ways this may not work...
        # TODO: cropping is likely done elsewhere
        crop_to_fit_region = True

        # Find the center of mass of the image closest to Z=0 and use it to
        # center the fit crop area for all images.
        z_zero_ind = find_closest(zarr, 0)
        img = np.array(Image.open(file_names[z_zero_ind]))

        # If the image is RGB, use the red channel
        # TODO: probably grayscale would be better
        imgsz = img.shape
        if len(imgsz) > 2 and img.shape[2] > 1:
            img = img[:, :, 0]

        # Background illumination can skew the center of mass of the image.
        # Set pixels at half the maximum intensity to zero to avoid that.
        # This might seem overzealous, but we are only masking here for center
        # of mass estimation, fitting will use all pixels and estimate the
        # background properly.
        img[img < np.max(img)/2] = 0
        fit_area_center = center_of_mass(img)

        # TODO: fit area is hardcoded, works well for ICX445, but perhaps not
        # for other sensors.
        fit_area_sz = [100, 100]
        fit_crop_area = np.round([
            fit_area_center[1] - fit_area_sz[1]/2,
            fit_area_center[1] + fit_area_sz[1]/2,
            fit_area_center[0] - fit_area_sz[0]/2,
            fit_area_center[0] + fit_area_sz[0]/2]).astype('int')


        for ind, file_name in enumerate(file_names):
            print("Processing file {:d} {:s}".format(ind, file_name))

            img = np.array(Image.open(file_name))
            imgsz = img.shape
            xarr = np.arange(0, imgsz[1])*pxsz
            yarr = np.arange(0, imgsz[0])*pxsz

            # Select the R channel in RGB images
            if len(imgsz)>2 and imgsz[2] > 1:
                img = img[:, : ,0]

            # Determine maximum intensity
            ampl[ind] = np.max(img)

            # Find center of mass
            center[ind, :] = center_of_mass(img)

            try:
                plt.figure()
                fit_result = fit_xy_profile(
                    img, X=xarr, Y=yarr, pxsz=pxsz, crop_area=fit_crop_area,
                    plot=True, suptitle_str="XY Profile at Z={:.2f}".format(
                        zarr[ind]))

                fit_arr[ind, 0] = zarr[ind]
                fit_arr[ind, 1:] = fit_result

                np.savetxt(get_default_overlap_fit_result_filename(), fit_arr)

                fig_file_name = "XY_profile_{:.2f}.png".format(zarr[ind])
                plt.draw()
                export_figure(fig_file_name)
                plt.close()

            except RuntimeWarning:
                # If there's a runtime warning, set fit values to NaN and move
                # on to other images
                print("A runtime warning occurred when processing "
                      "file {:s}".format(file_name))
                fit_result = [np.nan, np.nan]
            except Exception as excpt:
                # If there's an expection, set fit values to NaN as well and
                # move on to other images
                print("Could not fit file {:s}".format(file_name))
                print("Reason: ", excpt)
                fit_result = [np.nan, np.nan]

            width[ind, :] = fit_result[3:5]

        os.chdir(orig_path)

        # Calculate center positions in um
        center[:, 0] = center[:, 0]*pxsz
        center[:, 1] = center[:, 1]*pxsz

        center = fit_arr[:, 2:4]

    except Exception as excpt:
        os.chdir(orig_path)
        raise RuntimeError("Could not get overlap parameters") from excpt

    return [zarr, ampl, center, width]

def fit_xy_profile(data, **kwargs):
    """Fit a 2D gaussian to an XY profile."""
    return fit_gaussian_2d(data, **kwargs, return_fwhm=True)

def find_focus_position(
        zarr=None, ampl=None, color=None, yl=None, focus_fit_range=600,
        mode='max', fit_focus_curve=True, show_focus_marker=False, plot=True):
    """Find focus position.

    The function starts with a Z position estimate based on the maximum (or
    minimum) focus curve value. The focus curve is then fit with a parabola
    around the estimated position for a more precise result.

    Mode can be switched from 'max' to 'min' to facilitate cases where the
    signal is maximized and minimized at focus.
    """
    if mode=='max':
        z_focus_est = zarr[np.argmax(ampl)]
    elif mode=='min':
        z_focus_est = zarr[np.argmin(ampl)]

    # print.info("Estimated Z focus position: {:.1f} mm".format(z_focus_est))

    if z_focus_est > np.max(zarr) or z_focus_est < np.min(zarr):
        # TODO: Can this ever be true when using max/min functions above?
        zarr_focus = zarr
        ampl_focus = ampl
    else:
        [zarr_focus, ampl_focus] = cut_by_x_range(
            zarr, ampl,
            rng_fr=z_focus_est - focus_fit_range/2,
            rng_to=z_focus_est + focus_fit_range/2)

    if len(zarr_focus) < 3:
        print("Warning: not enough points (N<3) to find focal position")

    if fit_focus_curve:
        try:
            popt = fit_poly_2(
                X=zarr_focus,Y=ampl_focus, plot_fit=False, color=color)[0]
            z_focus = get_poly_2_max_x(popt)

            if plot:
                zarr_fmod = np.linspace(zarr.min(), zarr.max(), 100)
                plt.plot(
                    zarr_fmod, poly_2(zarr_fmod, *popt), color=color,ls='--')
                if show_focus_marker:
                    plt.plot([z_focus, z_focus], yl, color=color,ls='--')
        except Exception:
            print("Failed to fit the focus curve, using rought estimate as the"
                  " focus position")
            z_focus = z_focus_est
    else:
        z_focus = z_focus_est

    return z_focus

def report_num_failed_pts(fail_mask, message):
    """Print mask check result.

    Report the number of failed elements in an a mask check array.
    """
    num_failed = np.nansum(fail_mask)
    num_pts = len(fail_mask)
    if num_failed > 0:
        print("WARNING: {:d} out of {:d} ".format(num_failed, num_pts),
              message)

    # TODO: Might be useful to report the failed z positions too

def validate_profile_fit_results(par):
    """Validate beam profile fit results.

    Check whether the beam profile fit values are within the expected range.
    """
    # The maximum fit amplitude can be larger than 255 even though the 8-bit
    # data is capped at 255. Gaussian fits with narrow widths can spike well
    # above this value and still be valid. The 300 value is an educated guess.
    min_ampl = 10
    max_ampl = 300

    if par is None:
        return None

    name = par['name']

    ampl_fail = np.logical_or(par['ampl'] > max_ampl, par['ampl'] < min_ampl)
    report_num_failed_pts(ampl_fail, name + " fit amplitudes are "
                       "outside of the valid range")

    # Count the number of failed fits
    failed_pts = np.logical_or(np.logical_or(
            np.isnan(par['zpos']), np.isnan(par['width'][:, 0])), ampl_fail)

    if failed_pts.all():
        print("ERROR: All " + name + " fit values failed, there is no valid "
                    "data")
        return None
    if failed_pts.any():
        report_num_failed_pts(failed_pts, name + " fits failed, removing...")
        par['zpos'] = par['zpos'][np.logical_not(failed_pts)]
        par['ampl'] = par['ampl'][np.logical_not(failed_pts)]
        par['xypos'] = par['xypos'][np.logical_not(failed_pts), :]
        par['width'] = par['width'][np.logical_not(failed_pts), :]

    return par

def calc_plot_limits(par):
    """Calculate plot ranges for XYZ positions and XY widths."""
    zpos = par['zpos']
    xypos = par['xypos']
    width = par['width']

    # Determine XY position spans
    xspan = np.max(xypos[:,0]) - np.min(xypos[:,0])
    yspan = np.max(xypos[:,1]) - np.min(xypos[:,1])

    # Determine avg. width span
    width_avg = np.mean(width, 1)
    wspan = np.max(width_avg) - np.min(width_avg)

    # Calculate plot ranges for XYZ positions anx XY widths
    plot_xl = [np.min(xypos[:,0]) - xspan*0.1, np.max(xypos[:,0]) + xspan*0.1]
    plot_yl = [np.min(xypos[:,1]) - yspan*0.1, np.max(xypos[:,1]) + yspan*0.1]
    plot_zl = [np.min(zpos), np.max(zpos)]
    plot_wl = [ np.min(width_avg) - wspan*0.1, np.max(width_avg) + wspan*0.1 ]

    return [plot_xl, plot_yl, plot_zl, plot_wl]

def plot_overlap_scan_results(**kwargs):
    """Wrapper for old function name.

    TODO: Deprecate after 2025 Jan.
    """
    warn("Use plot_overlap_results() instead", DeprecationWarning)
    return plot_overlap_results(**kwargs)

def plot_overlap_results(
    zarr=None, ampl=None, center=None, width=None, width_metric='mean', show_focus_marker=True,
    ref_focus=None, z_axes=None, x_axes=None, y_axes=None, xy_axes=None,
    w_axes=None, xl=None, yl=None, zl=None, wl=None, color=None, show_fill=True, fill_color=None,
    with_titles=True, show_x_label=True, z_axes_ylabel_preffix=None, ampl_units=None,
    **kwargs):
    """Plot overlap analysis result figure.

    The figure contains three panels in which two curves show the overlap
    between pump and probe beams. Pump is blue, probe is red. The first panel
    shows the width of the beams as they go through focus. The second and third
    panels show the X and Y distances between the pump and probe beams as a
    function of Z.
    """
    if ref_focus is not None:
        focus_marker_color = get_color('dr')
    else:
        focus_marker_color = color

    xpos = center[:, 0]
    ypos = center[:, 1]


    ## === Plot width panel ===

    plt.sca(w_axes)

    # Plotting two (XY) curves for both beams makes the panel too cluttered.
    # Reduce the XY curves to one per beam using a given width metric for
    # better legibility.
    if width_metric not in ['mean', 'min', 'max']:
        warn('Undefined width mean metric, using mean')
        width_metric = 'mean'
    if width_metric == 'mean':
        width_plot = np.mean(width, 1)
    elif width_metric == 'min':
        width_plot = np.min(width, 1)
    elif width_metric == 'max':
        width_plot = np.max(width, 1)

    if show_fill:
        # Show width metric curve together with a fill to indicate the XY value
        # range which is useful for beam ellipticity
        plt.fill_between(zarr, width[:,0], width[:,1], color=fill_color)

    plt.plot(zarr, width_plot, c=color, linestyle='solid', marker='.')

    # Find the Z focus position from the width data
    z_focus = find_focus_position(
        zarr=zarr, ampl=width_plot, color=color, yl=[0,1.1], mode='min',
        focus_fit_range=2000, plot=False, **kwargs)

    # If no reference focus position is given, use the Z focus
    if ref_focus is None:
        ref_focus = z_focus

    # Show focus marker for reference
    if show_focus_marker:
        plt.plot([ref_focus, ref_focus], wl, c=focus_marker_color, ls='--')

    # Determine the average beam width at the reference focus position
    w_at_focus = np.interp(ref_focus, zarr, width_plot)
    plt.plot(zl, [w_at_focus, w_at_focus], c=color, ls='--')
    if with_titles:
        plt.title('FWHM={:.1f} µm'.format(w_at_focus))

    if show_x_label:
        plt.xlabel('Z position, µm')
    plt.ylabel('FWHM, µm ')
    plt.xlim(zl)
    plt.ylim(wl)
    plt.grid('on')
    add_x_marker(0, ls='-')
    plt.draw()


    ## === Plot X position panel ===

    plt.sca(x_axes)
    plt.plot(zarr, xpos, c=color, marker='.')

    # Mark the reference focus position
    plt.plot([ref_focus, ref_focus], xl, c=focus_marker_color,ls='--')

    # Determine the X position at the reference focus
    x_at_focus = np.interp(ref_focus, zarr, xpos)
    plt.plot(zl, [x_at_focus, x_at_focus], c=color, ls='--')
    if with_titles:
        plt.title('Focus position at X={:.1f} µm'.format(x_at_focus))

    if show_x_label:
        plt.xlabel('Z position, µm')
    plt.ylabel('X position, µm ')
    plt.xlim(zl)
    plt.ylim(xl)
    plt.grid('on')
    add_x_marker(0, ls='-')
    add_y_marker(0, ls='-')
    plt.draw()


    ## === Plot Y position panel ===

    plt.sca(y_axes)
    plt.plot(zarr, ypos, c=color, marker='.')

    # Mark the reference focus position
    plt.plot([ref_focus, ref_focus], yl, c=focus_marker_color, ls='--')

    # Determine the Y position at the reference focus
    y_at_focus = np.interp(ref_focus, zarr, ypos)
    plt.plot(zl,[y_at_focus,y_at_focus],c=color,ls='--')
    if with_titles:
        plt.title('Focus position at Y={:.1f} µm'.format(y_at_focus))

    if show_x_label:
        plt.xlabel('Z position, µm')
    plt.ylabel('Y position, µm ')
    plt.xlim(zl)
    plt.ylim(yl)
    plt.grid('on')
    add_x_marker(0, ls='-')
    add_y_marker(0, ls='-')
    plt.draw()

    return [z_focus, w_at_focus, x_at_focus, y_at_focus]

def make_xy_profile_gif(path='.', output_name="XYProfiles.gif", verbose=False):
    """Make a GIF of XY profile fit figures ordered by Z position."""
    output_path = join_paths(path, output_name)
    if check_file_exists(output_path):
        print("{:} already exists, assuming it is up to date and skipping GIF "
              "generation".format(output_path))
        return

    file_names = enum_files(
        path=path, extension="png", preffix="XY_profile", verbose=verbose,
        prepend_path=True)[0]

    if verbose:
        print("Found {:d} files".format(len(file_names)))
        print("Making GIF...")

    make_gif(file_names=file_names, output_name=output_path)

    if verbose:
        print("Done")

def make_overlap_report(**kwargs):
    """Wrapper for generating pump-probe overlap report with logging to file."""
    with LogPrinter():
        _make_overlap_report(**kwargs)

def _make_overlap_report(
        path=r'./', width_metric='mean', xy_pos_rel_to_z0=True,
        fig_style='whole_range', subplot_rows=1, subplot_row=0,
        plot_wl=None, suptitle_str=None, obj_id=None,
        device_sn=None, date_str=None):
    """Analyze beam overlap, make a report figure and animations."""
    if obj_id is None or device_sn is None:
        obj_id, device_sn = get_report_metainfo()

    # Number of overlap-related warnings
    overlap_warn = 0

    # Is the overlap result within specification
    overlap_in_spec = True

    # Supported objectives
    supported_obj = ['nikon_pf_4x', 'nikon_pf_10x']
    if obj_id not in supported_obj:
        warn("Objective '" + obj_id + "' is not supported, not all beam size "
             "checks will be performed")
        overlap_warn += 1
        obj_ind = 0
    else:
        for ind, name in enumerate(supported_obj):
            if name == obj_id:
                obj_ind = ind + 1
                break

    # Expected pump and probe beam size ranges
    # Objective indices: unknown, Plan Fluor 4x, Plan Fluor 10x
    min_probe_focus_sz = [0.1, 7, 3][obj_ind]
    max_probe_focus_sz = [50, 12, 12][obj_ind]
    max_probe_sz = [500, 100, 100][obj_ind]
    min_probe_focus_z = [-2, -0.1, -0.1][obj_ind]
    max_probe_focus_z = [2, 0.1, 0.1][obj_ind]

    min_pump_focus_sz = [0.1, 5, 3][obj_ind]
    max_pump_focus_sz = [50, 25, 25][obj_ind]
    max_pump_sz = [500, 100, 100][obj_ind]
    min_pump_focus_z = [-2, -0.5, -0.5][obj_ind]
    max_pump_focus_z = [2, 0.5, 0.5][obj_ind]

    # Allowed pump-probe focus lateral mismatch
    max_pp_xy_r = [50, 5, 4][obj_ind]

    # Allowed pump-probe focal spot size ratio
    min_overlap_ratio = 1.4
    max_overlap_ratio = 2.1

    path = Path(path)
    path_pu = path / 'Pump'
    path_pr = path / 'Probe'

    print("Working directory: {}".format(path))

    try:
        print("Getting pump focus parameters...")
        [zarr_pu, ampl_pu, xypos_pu, width_pu] = \
            get_xyz_overlap_parameters(path=path_pu)
        par_pu = {'name': 'pump', 'zpos': zarr_pu, 'ampl': ampl_pu,
                  'xypos': xypos_pu, 'width': width_pu}
    except Exception as excpt:
        warn("Pump data not available, pump metrics will be skipped. "
             "Reason: ", excpt)
        overlap_in_spec = False
        par_pu = None

    try:
        print("Getting probe focus parameters...")
        [zarr_pr, ampl_pr, xypos_pr, width_pr] = \
            get_xyz_overlap_parameters(path=path_pr)
        par_pr = {'name': 'probe', 'zpos': zarr_pr, 'ampl': ampl_pr,
                  'xypos': xypos_pr, 'width': width_pr}
    except Exception as excpt:
        warn("Probe data not available, probe metrics will be skipped. "
             "Reason: ", excpt)
        overlap_in_spec = False
        par_pu = None

    # Make sure either pump or probe data is available
    if par_pu is None and par_pr is None:
        raise RuntimeError("Neither pump nor probe data is available")

    if par_pu is not None and par_pr is not None:
        len_pu = len(par_pu['zpos'])
        len_pr = len(par_pr['zpos'])
        if len_pu != len_pr:
            raise RuntimeError("The number of pump ({:d})".format(len_pu) +
                  " and probe ({:d}) imges is not equal".format(len_pr))
    elif par_pu is not None:
        num_pts = len(par_pu['zpos'])
    else:
        num_pts = len(par_pr['zpos'])

    par_pu = validate_profile_fit_results(par_pu)
    par_pr = validate_profile_fit_results(par_pr)

    if par_pr is not None:
        beam_sz_pr = np.sqrt(np.mean(par_pr['width']**2, 1))

        probe_focus_sz = np.min(beam_sz_pr)
        if probe_focus_sz < min_probe_focus_sz:
            print("WARNING: Probe focus size is too small " +
                  "(<{:.1f} µm)".format(min_probe_focus_sz))
            overlap_warn += 1
        if probe_focus_sz > max_probe_focus_sz:
            print("WARNING: Probe focus size is too large " +
                  "(>{:.0f} µm)".format(max_probe_focus_sz))
            overlap_warn += 1

        probe_sz_max = np.max(beam_sz_pr)
        if probe_sz_max > max_probe_sz:
            print("WARNING: Maximum probe beam size is too large " +
                  "(>{:.0f} µm)".format(max_probe_sz))
            overlap_warn += 1

        probe_focus_ind = np.argmin(beam_sz_pr)
        probe_focus_x = xypos_pr[probe_focus_ind, 0]
        probe_focus_y = xypos_pr[probe_focus_ind, 1]

        probe_focus_z = zarr_pr[probe_focus_ind]
        if probe_focus_z < min_probe_focus_z:
            print("WARNING: Probe focus Z position is too close " +
                  "(<{:.1f} mm)".format(min_probe_focus_z))
            overlap_warn += 1
        if probe_focus_z > max_probe_focus_z:
            print("WARNING: Probe focus Z position is too far " +
                  "(>{:.1f} mm)".format(max_probe_focus_z))
            overlap_warn += 1

        print("Probe focus size is {:.1f} µm FWHM at Z = {:.2f} mm".format(
            probe_focus_sz, probe_focus_z))

    if par_pu is not None:
        beam_sz_pu = np.sqrt(np.mean(width_pu**2, 1))

        pump_focus_sz = np.min(beam_sz_pu)
        if pump_focus_sz < min_pump_focus_sz:
            print("WARNING: Pump focus size is too small " +
                  "(<{:.1f} µm)".format(min_pump_focus_sz))
            overlap_warn += 1
        if pump_focus_sz > max_pump_focus_sz:
            print("WARNING: Pump focus size is too large " +
                  "(>{:.0f} µm)".format(max_pump_focus_sz))
            overlap_warn += 1
            overlap_in_spec = False

        pump_sz_max = np.max(beam_sz_pu)
        if pump_sz_max > max_pump_sz:
            print("WARNING: Maximum pump beam size is too large " +
                  "(>{:.0f} µm)".format(max_pump_sz))
            overlap_warn += 1

        pump_focus_ind = np.argmin(beam_sz_pu)
        pump_focus_x = xypos_pu[pump_focus_ind, 0]
        pump_focus_y = xypos_pu[pump_focus_ind, 1]

        pump_focus_z = zarr_pu[pump_focus_ind]
        if pump_focus_z < min_pump_focus_z:
            print("WARNING: Pump focus Z position is too close " +
                  "(<{:.1f} mm)".format(min_pump_focus_z))
            overlap_warn += 1
        if pump_focus_z > max_pump_focus_z:
            print("WARNING: Pump focus Z position is too far " +
                  "(>{:.1f} mm)".format(max_pump_focus_z))
            overlap_warn += 1

        print("Pump focus size is {:.1f} µm FWHM at Z = {:.2f} mm".format(
            pump_focus_sz, pump_focus_z))

    # Calculate XY position with respect to probe focus
    if xy_pos_rel_to_z0 and par_pr is not None:
        par_pr['xypos'][:, 0] = par_pr['xypos'][:, 0] - probe_focus_x
        par_pr['xypos'][:, 1] = par_pr['xypos'][:, 1] - probe_focus_y

        if par_pu is not None:
            par_pu['xypos'][:, 0] = par_pu['xypos'][:, 0] - probe_focus_x
            par_pu['xypos'][:, 1] = par_pu['xypos'][:, 1] - probe_focus_y
            pump_focus_x -= probe_focus_x
            pump_focus_y -= probe_focus_y

        probe_focus_x = 0
        probe_focus_y = 0

    if par_pr is not None and par_pu is not None:
        pp_xy_r = np.sqrt((pump_focus_x - probe_focus_x)**2 + (pump_focus_y - probe_focus_y)**2)
        print("XY distance between pump and probe is {:.1f} µm".format(pp_xy_r))
        if pp_xy_r > max_pp_xy_r:
            print("WARNING: XY distance between pump and probe focus is too large " +
                  "(>{:.1f} µm)".format(max_pp_xy_r))
            overlap_warn += 1
            overlap_in_spec = False

    focus_region_size = 0.5
    if fig_style == 'at_focus':
        focus_rng = [probe_focus_z - focus_region_size/2, probe_focus_z + focus_region_size/2]
        if par_pr is not None:
            zpos = par_pr['zpos']
            par_pr['ampl'] = cut_by_x_range(zpos, par_pr['ampl'], rng=focus_rng)[1]
            par_pr['xypos'] = cut_by_x_range(zpos, par_pr['xypos'], rng=focus_rng)[1]
            par_pr['zpos'], par_pr['width'] = cut_by_x_range(zpos, par_pr['width'], rng=focus_rng)
        if par_pu is not None:
            zpos = par_pu['zpos']
            par_pu['ampl'] = cut_by_x_range(zpos, par_pu['ampl'], rng=focus_rng)[1]
            par_pu['xypos'] = cut_by_x_range(zpos, par_pu['xypos'], rng=focus_rng)[1]
            par_pu['zpos'], par_pu['width'] = cut_by_x_range(zpos, par_pu['width'], rng=focus_rng)

    if par_pu is not None:
        [plot_xl1, plot_yl1, plot_zl1, plot_wl1] = calc_plot_limits(par_pu)

    if par_pr is not None:
        [plot_xl2, plot_yl2, plot_zl2, plot_wl2] = calc_plot_limits(par_pr)

    if par_pu is not None:
        max_ampl_pu = np.max(par_pu['ampl'])
        par_pu['ampl'] = par_pu['ampl']/max_ampl_pu

    if par_pr is not None:
        max_ampl_pr = np.max(par_pr['ampl'])
        par_pr['ampl'] = par_pr['ampl']/max_ampl_pr

    ampl_units = 'a.u.'

    if par_pu is not None and par_pr is not None:
        xlims = [plot_xl1, plot_xl2]
        ylims = [plot_yl1, plot_yl2]
        zlims = [plot_zl1, plot_zl2]
        wlims = [plot_wl1, plot_wl2]
    elif par_pu is not None:
        xlims = [plot_xl1]
        ylims = [plot_yl1]
        zlims = [plot_zl1]
        wlims = [plot_wl1]
    elif par_pr is not None:
        xlims = [plot_xl2]
        ylims = [plot_yl2]
        zlims = [plot_zl2]
        wlims = [plot_wl2]

    plot_xl = get_common_range(xlims, mode='bound', expand_frac=0.1)
    plot_yl = get_common_range(ylims, mode='bound', expand_frac=0.1)
    if isnone(plot_wl):
        plot_wl = [0, 75]
        print("WARNING: Setting width axis range " +
                    "from {:.0f} µm ".format(plot_wl[0]) +
                    "to {:.0f} µm".format(plot_wl[1]))

        #plot_wl = get_common_range(wlims, mode='bound', expand_frac=0.1)
    plot_zl = get_common_range(zlims, mode='bound')

    plt.figure(figsize=[10, 8])
    subplot_col = 1
    subplot_cols = 3
    z_ax = None
    showxlabel=True

    w_ax = plt.subplot(subplot_rows, subplot_cols, subplot_col+subplot_row*subplot_cols)
    subplot_col = subplot_col+1
    x_ax = plt.subplot(subplot_rows, subplot_cols, subplot_col+subplot_row*subplot_cols)
    subplot_col = subplot_col+1
    y_ax = plt.subplot(subplot_rows, subplot_cols, subplot_col+subplot_row*subplot_cols)
    subplot_col = subplot_col+1

    # pump, probe
    colors = [get_color('db'), get_color('dr')]
    fill_colors = [get_color('b'), get_color('r')]

    if par_pr is not None:
        [z_focus_pr, w_at_focus_pr, x_at_focus_pr, y_at_focus_pr] = plot_overlap_results(par_pr['zpos'], par_pr['ampl'], par_pr['xypos'], par_pr['width'], \
            fit_focus_curve=False, \
            z_axes=z_ax, w_axes=w_ax, x_axes=x_ax, y_axes=y_ax, xl=plot_xl, yl=plot_yl, zl=plot_zl, wl=plot_wl,
            color=colors[1], fill_color=fill_colors[1],
            with_titles=False, show_x_label=showxlabel, ampl_units=ampl_units, width_metric=width_metric)
    if par_pu is not None:
        [z_focus_pu, w_at_focus_pu, x_at_focus_pu, y_at_focus_pu] = plot_overlap_results(par_pu['zpos'], par_pu['ampl'], par_pu['xypos'], par_pu['width'], \
            ref_focus=z_focus_pr, show_focus_marker=False, z_axes=z_ax, w_axes=w_ax, x_axes=x_ax, y_axes=y_ax, xl=plot_xl, yl=plot_yl, zl=plot_zl, wl=plot_wl,
            color=colors[0], fill_color=fill_colors[0],
            with_titles=False, show_x_label=showxlabel, ampl_units=ampl_units, width_metric=width_metric)

    if par_pu is not None and par_pr is not None:
        overlap_ratio = w_at_focus_pu/w_at_focus_pr

        if overlap_ratio < min_overlap_ratio:
            print("WARNING: Overlap ratio is too small, increase pump focus size")
            overlap_warn += 1
            overlap_in_spec = False
        if overlap_ratio > max_overlap_ratio:
            print("WARNING: Overlap ratio is too large, decrease pump focus size")
            overlap_warn += 1
            overlap_in_spec = False

    plt.sca(w_ax)
    overlap_str = ''
    if par_pu is not None and par_pr is not None:
        overlap_str += 'Pu/Pr FWHM overlap {:.1f}\n'.format(overlap_ratio)
    if par_pu is not None:
        overlap_str += 'Pump: {:.1f} µm'.format(w_at_focus_pu)
    if par_pr is not None:
        overlap_str += ' Probe: {:.1f} µm'.format(w_at_focus_pr)
    if not par_pu is not None or not par_pr is not None:
        overlap_str += ' FHWM'
    plt.title(overlap_str)

    plt.sca(x_ax)
    xofs_str = ''
    if par_pu is not None and par_pr is not None:
        xofs_str += 'X offset {:.1f} µm\n'.format(x_at_focus_pr - x_at_focus_pu)
    if par_pu is not None:
        xofs_str += 'Pump at {:.1f} µm'.format(x_at_focus_pu)
    if par_pr is not None:
        xofs_str += ' Probe at {:.1f} µm'.format(x_at_focus_pr)
    plt.title(xofs_str)

    plt.sca(y_ax)
    yofs_str = ''
    if par_pu is not None and par_pr is not None:
        yofs_str += 'Y offset {:.1f} µm\n'.format(y_at_focus_pr - y_at_focus_pu)
    if par_pu is not None:
        yofs_str += 'Pump at {:.1f} µm'.format(y_at_focus_pu)
    if par_pr is not None:
        yofs_str += ' Probe at {:.1f} µm'.format(y_at_focus_pr)
    plt.title(yofs_str)

    if suptitle_str is None:
        suptitle_str = "Overlap report"

    if device_sn is not None:
        suptitle_str += ", " + device_sn

    if obj_ind is not None:
        objective_str = ['unspecified objective', 'Nikon Plan Fluor 4x/0.13', 'Nikon Plan Fluor 10x/0.3'][obj_ind]
        suptitle_str += ", " + objective_str

    if date_str is None:
        date_str = time.strftime("%Y-%m-%d %H:%M")

    suptitle_str += ", " + date_str

    if overlap_in_spec:
        suptitle_str += "\n"
        suptitle_str += "Overlap is in spec"
    else:
        suptitle_str += "\n"
        suptitle_str += "Overlap is NOT IN SPEC, number of overlap-related warnings: {:d}".format(overlap_warn)

    plt.suptitle(suptitle_str)

    plt.gcf().set_size_inches(17, 5*subplot_rows + 1)
    if fig_style == 'whole_range':
        fig_name = 'Overlap_whole.pdf'
    if fig_style == 'at_focus':
        fig_name = 'Overlap_focus.pdf'

    plt.gcf().tight_layout(pad=3)

    export_figure(path / fig_name)

    print("Generating GIFs...")
    if par_pu is not None:
        try:
            make_xy_profile_gif(path=path_pu)
        except Exception:
            print("WARNING: Could not generate pump overlap GIF. " + get_exception_info_str())

    if par_pr is not None:
        try:
            make_xy_profile_gif(path=path_pr)
        except Exception:
            print("WARNING: Could not generate probe overlap GIF. " + get_exception_info_str())

    print("Showing report figure")

    plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""HARPIA Microscopy Module Python library.

Author: Lukas Kontenis
Copyright (c) 2019-2021 Light Conversion
All rights reserved.
www.lightcon.com
"""
import pkg_resources

spinnaker_ver = None
try:
    spinnaker_ver = pkg_resources.get_distribution('spinnaker-python').version
except pkg_resources.DistributionNotFound:
    print("Spinnaker API not installed")

if spinnaker_ver is not None:
    print("Loading Spinnaker API {:s}...".format(spinnaker_ver),
          end='', flush=True)
    import PySpin
    print('OK')

    system = PySpin.System.GetInstance()
    cam_list = system.GetCameras()

    if len(cam_list) > 0:
        print("Camera list:")
        for ind, cam in enumerate(cam_list):
            cam.Init()
            model_str = cam.DeviceModelName.GetValue()
            sn_str = cam.DeviceSerialNumber.GetValue()
            cam.DeInit()
            print("Cam {:d}, {:s}, SN: {:s}".format(ind, model_str, sn_str))
    else:
        print("No cameras found")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""HARPIA Microscopy Module Python library.

support@lightcon.com
www.lightcon.com
All rights reserved.
© 2019-2024 Light Conversion
"""
__version__ = '0.3.2'

# from . import harpiamm
# from . import harpiamm_ui
# from . import harpiamm_ui_cli
# from . import delay_line_alignment
# from . import cam_focus_test

# Qt GUI supports depends on PySide2
# from . import harpiamm_ui_qt



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""HARPIA-MM pump-probe XYZ overlap characterization using a camera.

This module provides routines for the characterization of the pump and probe
beam focus and overlap using a camera placed at the sample plane. Images are
acquired as the camera is moved though the focus by the sample stage. The
camera is controlled via the FLIR/Teledyne Spinnaker API, and the stage is
controlled using HARPIA API.

This script is a part of the HARPIA Microscopy Kit, which is a set of tools for
the alignment, characterization and troubleshooting of the HARPIA-MM extension
of the HARPIA spectroscopy system.

support@lightcon.com
www.lightcon.com
All rights reserved.
© 2019-2024 Light Conversion
"""
import time
import os
import json
import datetime
from pathlib import Path

import numpy as np
import matplotlib.pyplot as plt
from PIL import Image
import PySpin

from lightcon.harpia import Harpia


def read_setup(par_name=None, file_name='setup.xml'):
    """Read setup XML file."""
    try:
        with open(file_name, encoding="utf-8") as file:
            setup = json.load(file)
            return setup.get(par_name, None)
    except Exception:
        return None

def get_image_with_ae(cam, img_mask=None):
    """Get camera image with autoexposure."""

    # Make sure that the camera exposure is set in us
    exposure_units = cam.ExposureTime.GetUnit()
    if exposure_units == 'us':
        exposure_fac = 1E-6
    else:
        raise ValueError("Unsupported exposure time units '{:s}'".format(
            exposure_units))

    max_exposure = cam.ExposureTime.GetMax()
    min_exposure = cam.ExposureTime.GetMin()
    exposure = cam.ExposureTime.GetValue()

    while True:
        # First image in an acquisition can be stale. Make sure to clear it
        # before calculating autoexposure.
        cam.AcquisitionStart()
        image_result = cam.GetNextImage()
        image_result.Release()

        cam.AcquisitionStart()
        image_result = cam.GetNextImage()
        img = image_result.GetData()

        img = img.reshape(image_result.GetHeight(), image_result.GetWidth())
        img = np.array(img)
        if img_mask is not None:
            if img.shape == img_mask.shape:
                img[img_mask] = 0
            else:
                print("Incorrect mask size. Clearing mask")
                img_mask = None

        max_val = img.max()
        if max_val >= 255:
            new_exposure = exposure*0.9
            if new_exposure < min_exposure:
                print("Max image value: {:}. Minimum exposure reached, cannot "
                      "decrease.".format(max_val))
                return img

            exposure = new_exposure
            print("Max image value: {:}. Decreasing exposure "
                  "to: {:.2f} ms".format(max_val, exposure*exposure_fac*1E3))
            cam.ExposureTime.SetValue(exposure)
        elif max_val < 100:
            new_exposure = exposure*1.1
            if new_exposure > max_exposure:
                print("Max image value: {:}. Maximum exposure reached, cannot "
                      "increase.".format(max_val))
                return img

            exposure = new_exposure
            print("Max image value: {:}. Increasing exposure "
                  "to: {:.2f} ms".format(max_val, exposure*exposure_fac*1E3))
            cam.ExposureTime.SetValue(exposure)
        else:
            return img


def cam_focus_test(
        start_pos=-2, end_pos=2, num_steps=50,
        meas_seq='bmb',
        cam_id=None, cam_sn=None, show_img=False,
        measure_pump=True, measure_probe=True,
        use_cam=True, use_shutters=True, dry_run=False):
    """Scan HARPIA XYZ sample stage Z axis and take an image at each step.

    The scan extent is configured by '''start_pos''', and '''end_pos''' and
    '''num_steps'''. Start and stop positions are given in mm.

    Measurement sequence is determined by '''meas_seq'''.
    In beam-move-beam ('bmb') sequence, the probe shutter is open, images are
    acquired for each stage position, and the probe shutter is closed. The
    procedure is then repeated for the pump beam.

    In move-beam-beam ('mbb') sequence, the stage is stepped over measurement
    positions, and at each position two images are taken for the pump and the
    probe beams by opening and closing each shutter.

    The 'bmb' sequence is recommended because it has fewer shutter open and
    close events and is thus faster than 'mbb'.

    In 'mbb' sequence the pump and probe images are acquired one after the
    other without moving the sample stage, resulting in a lower time delay
    between the images. In some setups the probe beam needs time to stabilize,
    so the 'mbb' sequence becomes slow or prone to systemic errors if the probe
    beam measurement is taken too soon after the shutter is opened.

    The sample stage camera can be selected by ID as enumerated by the camera
    driver using '''cam_id''', or serial number using '''cam_sn'''.

    By default the script measures both pump and probe beams unless
    '''measure_pump''' or '''measure_probe''' are set to false.

    The use of a camera or shutters can be disabled for debugging via
    '''use_cam''' and '''use_shutters''', although this will likely produce
    unusable data.

    The script can also be run in dry mode using '''dry_run''' to make sure the
    measurement sequence works as intended without moving any motors.
    """
    print("=== HARPIA pump-probe XYZ overlap camera scan ===")

    if cam_sn is None:
        cam_sn = read_setup('cam_sn')
        if cam_sn is None:
            print("Sample plane camera SN not specified, using the first "
                  "available camera")
            cam_id = 0
        else:
            print("Sample plane camera SN: " + cam_sn)

    # Wait after shutter is open for the beam to stabilize
    probe_on_delay_s = 2
    pump_on_delay_s = 0.5

    print("Scan configuration: from {:.2f} mm, to {:.2f} mm, "
          "num_steps {:d}".format(start_pos, end_pos, num_steps))

    # Make sure that the data directory exists
    try:
        os.mkdir(r".//data")
    except FileExistsError:
        pass
    except Exception as excpt:
        raise RuntimeError("Cannot create data folder") from excpt

    # Make sure the script has write access to the data directory
    try:
        tmp_file_name = r".//data/test.tmp"
        np.savetxt(tmp_file_name, [123])
    except PermissionError as excpt:
        raise RuntimeError("Write access to the data folder is denied. Run the"
                           " script from a folder where the current user has "
                           "write access.") from excpt
    os.remove(tmp_file_name)

    # Create a timestsamped folder for the current scan
    timestamp = datetime.datetime.now()
    timestamp_str = timestamp.strftime('%Y-%m-%d %H%M%S')
    data_dir = 'data/' + timestamp_str + ' - Overlap scan/'
    os.mkdir(data_dir)

    # Create pump and probe subdirs
    pump_dir = data_dir + "Pump/"
    probe_dir = data_dir + "Probe/"
    os.mkdir(pump_dir)
    os.mkdir(probe_dir)

    # TODO: It would be better to abstract specific camera API or CameraApp
    # REST on a separate abstact camera layer (MSD-2228)
    if use_cam:
        # Init camera
        system = PySpin.System.GetInstance()
        cam_list = system.GetCameras()

        if len(cam_list) == 0:
            print("No cameras found. Cannot continue.")
            return

        cam = None
        if cam_id is not None:
            cam = cam_list[cam_id]
        else:
            flir_sn_vid_bug = False
            print("Number of cameras detected: {:d}".format(len(cam_list)))
            for cam_ind, cam2 in enumerate(cam_list):
                list_cam_sn = cam2.GetUniqueID()
                print("\t{:d}: {:s}".format(cam_ind, list_cam_sn))
                if list_cam_sn.find('VID_') != -1:
                    flir_sn_vid_bug = True
                if list_cam_sn == cam_sn:
                    cam = cam2

            if flir_sn_vid_bug:
                print("WARNING: USB VID returned instead of camera SN. This is"
                      " a known FLIR Spinaker API bug. The requested camera "
                      "will likely be not detected even if the provided SN is "
                      "corrected. You can pass an explicit camera index using "
                      "'cam_id' if you can recognize the camera you need in "
                      "the list above. If there are only two FLIR cameras, "
                      "e.g. a PC running only a HARPIA-MM pass cam_id=1 if the"
                      " default behaviour opens the incorrect camera. "
                      "RepluggingIf all cameras or rebooting the PC usually "
                      "fixes the issue and correct SN strings are reuturned.")

        if cam is None:
            cam = cam_list[0]
            print("WARNING: Camera with SN {:s} not found. Using the first "
                  "available camera with SN {:s}".format(
                      cam_sn, cam.GetUniqueID()))

        # Sample plane cameras get damaged often resulting in stuck pixels
        # that mess up autoexposure and beam shape fitting. An image mask
        # is used to suppress unwanted pixels by setting them to zero.
        img_mask = None
        mask_file_name = 'img_mask.png'
        if Path(mask_file_name).is_file():
            print("Loading image mask from '{:s}'".format(mask_file_name))
            mask_data = Image.open(mask_file_name)
            print("Mask image size is {:d}x{:d}, mode: {:s}".format(
                mask_data.size[0], mask_data.size[1], mask_data.mode))
            if mask_data.mode != '1':
                # The image mask is supposed to be binary, but sometimes it is
                # useful to just crop and save an acquired image. If the mask
                # is not binary, convert it to binary and save it.
                print("Converting mask to binary")
                mask_data.convert('1')
                mask_data.save(mask_file_name)
            img_mask = np.array(mask_data) != 0

    # Init completed, now the actual scan

    try:
        start_t = time.time()
        if use_cam:
            cam.Init()

            model_str = cam.DeviceModelName.GetValue()
            sn_str = cam.DeviceSerialNumber.GetValue()
            print("Connected to " + model_str + ', SN: ' + sn_str)

            # Set acquisition mode to single frame
            # NOTE: Continuous might sound better, but there are some strange
            # buffering issues, which result in old images being returned as
            # the sample stage is moved.
            nodemap = cam.GetNodeMap()
            node_acquisition_mode = PySpin.CEnumerationPtr(
                nodemap.GetNode('AcquisitionMode'))

            node_acquisition_mode_singleframe = \
                node_acquisition_mode.GetEntryByName('SingleFrame')

            acq_mode_singleframe = node_acquisition_mode_singleframe.GetValue()
            node_acquisition_mode.SetIntValue(acq_mode_singleframe)


            # Turn off auto exposure and auto gain
            # TODO: We probably want to reset these parameters to what they
            # were after the script is completed
            cam.ExposureAuto.SetValue(0)
            cam.GainAuto.SetValue(0)

            cam.BeginAcquisition()

        pos_arr = np.linspace(start_pos, end_pos, num_steps)

        # Connect to a local HARPIA
        harpia_ip = '127.0.0.1'
        print("Connecting to HARPIA Service App at " + harpia_ip)
        try:
            harpia = Harpia(harpia_ip)
        except Exception as excpt:
            raise RuntimeError("Could not connect to HARPIA Service App") \
                from excpt

        if use_shutters:
            if measure_probe and not dry_run:
                harpia.close_probe_shutter()
            if measure_pump and not dry_run:
                harpia.close_pump_shutter()

        if show_img:
            plt.figure(1)
            plt.clf()

        if measure_probe and measure_pump:
            probe_ax = plt.subplot(1, 2, 1)
            pump_ax = plt.subplot(1, 2, 2)
        else:
            if measure_probe:
                probe_ax = plt.gca()
            if measure_pump:
                pump_ax = plt.gca()

        default_exp = 8000

        if use_cam:
            cam.ExposureTime.SetValue(default_exp)

        pump_exp = default_exp
        probe_exp = default_exp

        if meas_seq == 'mbb':
            # MBB strategy: move, measure probe and pump, then move
            for ind,pos in enumerate(pos_arr):
                print("Step {:} of {:}, pos={:.2f} mm".format(
                    ind+1, num_steps, pos))
                if not dry_run:
                    harpia.microscope_set_sample_stage_position_Z(pos*1000)

                if measure_probe:
                    # Measure probe beam
                    if use_shutters and not dry_run:
                        harpia.open_probe_shutter()
                        time.sleep(probe_on_delay_s)

                    if use_cam:
                        cam.ExposureTime.SetValue(probe_exp)

                        print("Acquiring probe image...")
                        img = get_image_with_ae(cam, img_mask=img_mask)
                        print("Done. Image val range [{:d}, {:d}]".format(
                            np.min(img), np.max(img)))
                        probe_exp = cam.ExposureTime.GetValue()
                    if use_shutters and not dry_run:
                        harpia.close_probe_shutter()

                    if use_cam:
                        plt.imsave(probe_dir + "img_{:.2f}.png".format(
                            pos_arr[ind]), img, cmap='gray', vmin=0, vmax=255)

                    if show_img:
                        plt.sca(probe_ax)
                        plt.imshow(img)

                if measure_pump:
                    # Measure pump beam
                    if use_shutters and not dry_run:
                        harpia.open_pump_shutter()
                        time.sleep(pump_on_delay_s)

                    if use_cam:
                        cam.ExposureTime.SetValue(pump_exp)
                        print("Acquiring pump image...")
                        img = get_image_with_ae(cam, img_mask=img_mask)
                        print("Done. Image val range [{:d}, {:d}]".format(
                            np.min(img), np.max(img)))
                        pump_exp = cam.ExposureTime.GetValue()

                    if use_shutters and not dry_run:
                        harpia.close_pump_shutter()

                    if use_cam:
                        plt.imsave(pump_dir + "img_{:.2f}.png".format(
                            pos_arr[ind]), img, cmap='gray', vmin=0, vmax=255)

                    if show_img:
                        plt.sca(pump_ax)
                        plt.imshow(img)

                if show_img:
                    plt.pause(0.05)

        elif meas_seq == 'bmb':
            # BMB strategy: move and measure all probe positions, then pump
            if measure_probe:
                # Measure probe beam
                if use_shutters and not dry_run:
                    harpia.open_probe_shutter()
                    time.sleep(probe_on_delay_s)

                for ind, pos in enumerate(pos_arr):
                    print("Step {:} of {:}, pos={:.2f} mm".format(
                        ind+1, num_steps, pos))
                    if not dry_run:
                        harpia.microscope_set_sample_stage_position_Z(pos*1000)

                    if use_cam:
                        cam.ExposureTime.SetValue(probe_exp)
                        print("Acquiring image...")
                        img = get_image_with_ae(cam, img_mask=img_mask)
                        print("Done. Image val range [{:d}, {:d}]".format(
                            np.min(img), np.max(img)))
                        probe_exp = cam.ExposureTime.GetValue()
                        plt.imsave(probe_dir + "img_{:.2f}.png".format(
                            pos_arr[ind]), img, cmap='gray', vmin=0, vmax=255)

                    if show_img:
                        plt.sca(probe_ax)
                        plt.imshow(img)
                        plt.pause(0.05)

            if measure_pump:
                # Measure pump beam
                if use_shutters and not dry_run:
                    harpia.close_probe_shutter()
                    harpia.open_pump_shutter()
                    time.sleep(pump_on_delay_s)

                for ind,pos in enumerate(pos_arr):
                    print("Step {:} of {:}, pos={:.2f} mm".format(
                        ind+1, num_steps, pos))
                    if not dry_run:
                        harpia.microscope_set_sample_stage_position_Z(pos*1000)
                    print("Move completed")

                    if use_cam:
                        cam.ExposureTime.SetValue(pump_exp)
                        print("Acquiring image...")
                        img = get_image_with_ae(cam, img_mask=img_mask)
                        print("Done. Image val range [{:d}, {:d}]".format(
                            np.min(img), np.max(img)))
                        pump_exp = cam.ExposureTime.GetValue()
                        plt.imsave(pump_dir + "img_{:.2f}.png".format(
                            pos_arr[ind]), img, cmap='gray', vmin=0, vmax=255)

                    if show_img:
                        plt.sca(pump_ax)
                        plt.imshow(img)
                        plt.pause(0.05)

        if use_shutters and not dry_run:
            if measure_probe:
                harpia.close_probe_shutter()
            if measure_pump:
                harpia.close_pump_shutter()

        print("Moving to starting position: {:.2f} mm".format(pos_arr[0]))
        if not dry_run:
            harpia.microscope_set_sample_stage_position_Z(pos_arr[0]*1000)

        if use_cam:
            cam.EndAcquisition()
            cam.DeInit()
            del cam

        print("All done")
    except Exception as excpt:
        raise RuntimeError("Scanning failed") from excpt

    print("Total time elapsed: {:.1f} s".format(time.time() - start_t))


if __name__ == '__main__':
    cam_focus_test()



"""HARPIA Microscopy Module Python library.

This module contains functions for testing pump-probe measurement sensitivity
in terms of the minimum resolvable differential optical density (ΔOD or DOD).
These functions were developed as part of the HARPIA Microscopy Module white
light supercontinuum prboe characterization when it is tightly focused using a
microscope objective.

Contact: lukas.kontenis@lightcon.com, support@lightcon.com

Copyright (c) 2019-2023 Light Conversion
All rights reserved.
www.lightcon.com
"""
import json

import matplotlib.pyplot as plt
import numpy as np

from lkcom.plot import add_y_marker
from lkcom.util import find_closest


def plot_dod_hist(
        dod_arr=None, wavl=None, wavl_rng=None,
        show_labels=False, show_title=True, ylim=None,
        show_mod_units_in_legend=False, log_hist=True,
        hist_range=[-2, 2]):
    """Plot a DOD histogram."""
    cut_ind = find_closest(wavl, wavl_rng)
    hist_data = np.array(dod_arr)[:, cut_ind[0]:cut_ind[1]].flatten()*1E3
    plt.hist(hist_data, bins=64, range=hist_range, log=log_hist)
    plt.grid()

    plt.xlim(hist_range)
    plt.ylim(ylim)

    if log_hist:
        y_pos_ofs_fac = 0.50
    else:
        y_pos_ofs_fac = 0.10

    xlim = plt.xlim()
    ylim = plt.ylim()
    xpos = xlim[1] - 0.05*(xlim[1]-xlim[0])
    ypos = ylim[1] - y_pos_ofs_fac*(ylim[1]-ylim[0])

    wavl_rng_str = "{:.0f} – {:.0f} nm".format(wavl_rng[0], wavl_rng[1])

    info_str = ''
    if not show_title:
        info_str = wavl_rng_str + '\n'

    if show_mod_units_in_legend:
        unit_str = 'mOD'
    else:
        unit_str = ''

    info_str += \
        'q_50: {:.3f} {:s}\n'.format(
            np.percentile(hist_data, 75) - np.percentile(hist_data, 25),
            unit_str) + \
        'q_90: {:.3f} {:s}'.format(
            np.percentile(hist_data, 95) - np.percentile(hist_data, 5),
            unit_str)

    plt.text(
        xpos, ypos, info_str, ha='right', va='top',
        bbox={'facecolor': 'white', 'edgecolor': 'none', 'alpha': 0.75})

    if show_title:
        plt.title(wavl_rng_str)

    if show_labels:
        plt.xlabel('Sen1sitivity, mOD')
        plt.ylabel('Counts')


def make_dod_sens_hist_fig(data_dir='./'):
    """Make a DOD sensitivity figure with histograms.

    Make a figure showing DOD sensitivity which contains two panels:
        1) Sample probe spectrum and DOD spectra at different ideal chop rates
        2) DOD histograms at a single chop rate split into wavelength bins
    """
    # === Load configuration ===
    config = json.load(open(data_dir + 'config.xml'))
    timestamp_str = config.get('timestamp_str')
    title_str = config.get('title_str', 'WLSC stability')
    num_acq_spectra = config.get('num_acq_spectra')
    num_spec_px = config.get('num_spec_px', 256)
    spectrum_time = config.get('spectrum_time')

    chop_bins = config.get('chop_bins', [1, 5, 10, 20])

    dod_wavl_arr = config.get(
        'dod_wavl_arr',
        [[510, 600], [600, 700], [700, 800], [800, 900]])

    wavl = np.loadtxt(data_dir + 'wavl.dat')
    spec = np.loadtxt(data_dir + 'spec.dat')
    dod_arr = np.loadtxt(data_dir + 'dod_arr.dat')*1E-3

    # === Parse data ===
    if timestamp_str is not None:
        title_str += ', ' + timestamp_str

    pump_spec = np.ndarray([int(num_acq_spectra/2), num_spec_px])
    unpump_spec = np.ndarray([int(num_acq_spectra/2), num_spec_px])

    legend_str = []
    for chop_bin in chop_bins:
        legend_str.append("{:.0f} Hz".format(np.round(1/(spectrum_time*chop_bin))))

    fig = plt.figure(figsize=[10, 3.5])
    gs = fig.add_gridspec(4, 2)
    ax_dod_spec_at_chop = plt.subplot(gs[:, 0])

    # === DOD spectra at several chop rates ===
    #
    # Ideal chopping is when the acquired spectra are sliced into blocks of N
    # spectra, i.e. the slices are always of the same size and there are no
    # pumped/unpumped spectra uncertainty.
    plt.subplot(ax_dod_spec_at_chop)
    plt.cla()
    chop_freq = 1/(np.array(chop_bins)*spectrum_time*2)
    legend_str = []
    for ind_chop, chop_bin in enumerate(chop_bins):
        for ind in np.arange(0, num_acq_spectra, chop_bin*2):
            hind = int(ind/2)
            pump_spec[hind:hind+chop_bin, :] = spec[ind:ind+chop_bin, :]
            unpump_spec[hind:hind+chop_bin, :] = spec[ind+chop_bin:ind+2*chop_bin, :]

        dod = np.log10(np.mean(unpump_spec, 0)/np.mean(pump_spec, 0))

        plt.plot(wavl, dod*1E3)
        legend_str.append("{:.0f} Hz".format(chop_freq[ind_chop]))

    avg_probe = np.mean(spec, 0)
    avg_probe = 4*(avg_probe/10 - 0.5)
    plt.plot(wavl, avg_probe, '-k')

    plt.ylim(-2, 2)
    plt.xlim([wavl[0], wavl[-1]])
    plt.grid('on')

    add_y_marker(0)

    plt.ylabel('mOD')
    plt.xlabel('Wavelength, nm')
    plt.title('Probe and sample DOD spectra')
    plt.legend(legend_str)

    # === DOD stability histograms ===
    for ind, wavl_rng in enumerate(dod_wavl_arr):
        plt.subplot(gs[ind, 1])
        plt.cla()
        plot_dod_hist(dod_arr, wavl, dod_wavl_arr[ind], ylim=[1, 2000],
                      show_title=False)

        plt.grid('on', which='both', axis='y')
        plt.gca().set_yticklabels([])
        if ind != len(dod_wavl_arr) - 1:
            plt.gca().set_xticklabels([])
        if ind == 0:
            plt.title('DOD histograms (log)')

    plt.xlabel('mOD')

    gs.update(hspace=0)
    plt.tight_layout()
    plt.draw()
    plt.pause(0.001)

    plt.tight_layout()
    # plt.suptitle(title_str)
    plt.savefig(timestamp_str + '-dod_hist.png')

    plt.draw()
    plt.pause(0.001)

    # Print percentile ranges
    print("Percentile ranges: ")
    for ind, wavl_rng in enumerate(dod_wavl_arr):
        cut_ind = find_closest(wavl, wavl_rng)
        hist_data = np.array(dod_arr)[:, cut_ind[0]:cut_ind[1]].flatten()*1E3
        q_50_str = "{:.3f}".format(
            np.percentile(hist_data, 75) - np.percentile(hist_data, 25))
        q_90_str = "{:.3f}".format(
            np.percentile(hist_data, 95) - np.percentile(hist_data, 5))
        print(q_50_str + ', ', q_90_str + ',  ', end='')
    print('\n')

    print("Close figure to exit")
    plt.show()
